"""The pytest-celery plugin provides a set of built-in components called
:ref:`vendors`.

This module is part of the :ref:`built-in-worker` vendor.
"""

import os

CELERY_SETUP_WORKER = "celery_setup_worker"
DEFAULT_WORKER = "default_worker_container"
WORKER_DOCKERFILE_ROOTDIR = os.path.dirname(__file__)
WORKER_CELERY_APP_NAME = "celery_test_app"
WORKER_CELERY_VERSION = ""  # latest from pypi
WORKER_LOG_LEVEL = "INFO"
WORKER_NAME = "celery_test_worker"
WORKER_QUEUE = "celery"
WORKER_ENV = {
    "CELERY_BROKER_URL": "memory://",
    "CELERY_RESULT_BACKEND": "cache+memory://",
    "PYTHONUNBUFFERED": "1",
    "PYTHONDONTWRITEBYTECODE": "1",
    "PYTHONPATH": "/app",
}
WORKER_DEBUGPY_PORTS = {
    "5678/tcp": "5678",
}
WORKER_VOLUME = {
    "bind": "/app",
    "mode": "rw",
}
DEFAULT_WORKER_APP_NAME = WORKER_CELERY_APP_NAME
DEFAULT_WORKER_VERSION = WORKER_CELERY_VERSION
DEFAULT_WORKER_LOG_LEVEL = WORKER_LOG_LEVEL
DEFAULT_WORKER_NAME = WORKER_NAME
DEFAULT_WORKER_ENV = WORKER_ENV
DEFAULT_WORKER_PORTS = None
DEFAULT_WORKER_QUEUE = WORKER_QUEUE
DEFAULT_WORKER_CONTAINER_TIMEOUT = 60
DEFAULT_WORKER_VOLUME = WORKER_VOLUME



"""Every broker component is added to the test matrix using the fixtures of
this module.

These fixtures will configure the test setup for all supported celery
brokers by default. Every broker will be executed as a separate test
case, and the test will be executed for each supported celery broker.

You may override these fixtures to customize the test setup for your
specific needs.
"""

# mypy: disable-error-code="misc"

from __future__ import annotations

import pytest

from pytest_celery.api.broker import CeleryBrokerCluster
from pytest_celery.api.broker import CeleryTestBroker
from pytest_celery.defaults import ALL_CELERY_BROKERS
from pytest_celery.defaults import CELERY_BROKER_CLUSTER


@pytest.fixture(params=ALL_CELERY_BROKERS)
def celery_broker(request: pytest.FixtureRequest) -> CeleryTestBroker:  # type: ignore
    """Parameterized fixture for all supported celery brokers. Responsible for
    tearing down the node.

    This fixture will add parametrization to the test function, so that
    the test will be executed for each supported celery broker.
    """
    broker: CeleryTestBroker = request.getfixturevalue(request.param)
    yield broker
    broker.teardown()


@pytest.fixture
def celery_broker_cluster(celery_broker: CeleryTestBroker) -> CeleryBrokerCluster:  # type: ignore
    """Defines the cluster of broker nodes for the test. Responsible for
    tearing down the cluster.

    It is not recommended to disable the broker cluster, but it can be done by
    overriding this fixture and returning None.

    Args:
        celery_broker (CeleryTestBroker): Parameterized fixture for all supported celery brokers.

    Returns:
        CeleryBrokerCluster: Single node cluster for all supported celery brokers.
    """
    cluster = CeleryBrokerCluster(celery_broker)  # type: ignore
    yield cluster
    cluster.teardown()


@pytest.fixture
def celery_broker_cluster_config(request: pytest.FixtureRequest) -> dict | None:
    """Attempts to compile the celery configuration from the cluster."""
    try:
        use_default_config = pytest.fail.Exception
        cluster: CeleryBrokerCluster = request.getfixturevalue(CELERY_BROKER_CLUSTER)
        return cluster.config() if cluster else None
    except use_default_config:
        return CeleryBrokerCluster.default_config()



"""The pytest-celery plugin provides a set of built-in components called
:ref:`vendors`. This module is part of the :ref:`built-in-worker` vendor.

Template for Celery worker application.
"""

from __future__ import annotations

import json

from celery import Celery

imports = None

app = Celery("celery_test_app")
config = None

if config:
    app.config_from_object(config)
    print(f"Changed worker configuration: {json.dumps(config, indent=4)}")


if __name__ == "__main__":
    app.start()



"""The pytest-celery plugin provides a set of built-in components called
:ref:`vendors`.

This module is part of the Memcached Backend vendor.
"""

# mypy: disable-error-code="misc"

from __future__ import annotations

import pytest
from pytest_docker_tools import container
from pytest_docker_tools import fxtr

from pytest_celery.vendors.memcached.api import MemcachedTestBackend
from pytest_celery.vendors.memcached.container import MemcachedContainer
from pytest_celery.vendors.memcached.defaults import MEMCACHED_CONTAINER_TIMEOUT


@pytest.fixture
def celery_memcached_backend(default_memcached_backend: MemcachedContainer) -> MemcachedTestBackend:
    """Creates a MemcachedTestBackend instance. Responsible for tearing down
    the node.

    Args:
        default_memcached_backend (MemcachedContainer): Instantiated MemcachedContainer.
    """
    backend = MemcachedTestBackend(default_memcached_backend)
    yield backend
    backend.teardown()


@pytest.fixture
def default_memcached_backend_cls() -> type[MemcachedContainer]:
    """Default Memcached backend container class. Override to apply custom
    configuration globally.

    See also: :ref:`vendor-class`.

    Returns:
        type[MemcachedContainer]: API for managing the vendor's container.
    """
    return MemcachedContainer


default_memcached_backend = container(
    image="{default_memcached_backend_image}",
    ports=fxtr("default_memcached_backend_ports"),
    environment=fxtr("default_memcached_backend_env"),
    network="{default_pytest_celery_network.name}",
    wrapper_class=MemcachedContainer,
    timeout=MEMCACHED_CONTAINER_TIMEOUT,
)


@pytest.fixture
def default_memcached_backend_env(default_memcached_backend_cls: type[MemcachedContainer]) -> dict:
    """Environment variables for this vendor.

    Args:
        default_memcached_backend_cls (type[MemcachedContainer]): See also: :ref:`vendor-class`.

    Returns:
        dict: Items to pass to the container's environment.
    """
    return default_memcached_backend_cls.initial_env()


@pytest.fixture
def default_memcached_backend_image(default_memcached_backend_cls: type[MemcachedContainer]) -> str:
    """Docker image for this vendor.

    Args:
        default_memcached_backend_cls (type[MemcachedContainer]): See also: :ref:`vendor-class`.

    Returns:
        str: Docker image name.
    """
    return default_memcached_backend_cls.image()


@pytest.fixture
def default_memcached_backend_ports(default_memcached_backend_cls: type[MemcachedContainer]) -> dict:
    """Port bindings for this vendor.

    Args:
        default_memcached_backend_cls (type[MemcachedContainer]): See also: :ref:`vendor-class`.

    Returns:
        dict: Port bindings.
    """
    return default_memcached_backend_cls.ports()



"""The pytest-celery plugin provides a set of built-in components called
:ref:`vendors`.

This module is part of the :ref:`built-in-worker` vendor.
"""

from __future__ import annotations

import json

import psutil


def get_running_processes_info(columns: list[str] | None = None) -> str:
    """Get information about running processes using psutil."""
    if not columns:
        columns = [
            "pid",
            "name",
            "username",
            "cmdline",
            "cpu_percent",
            "memory_percent",
            "create_time",
        ]
    processes = [proc.info for proc in psutil.process_iter(columns)]
    return json.dumps(processes)



"""The pytest-celery plugin provides a set of built-in components called
:ref:`vendors`.

This module is part of the Memcached Backend vendor.
"""

CELERY_MEMCACHED_BACKEND = "celery_memcached_backend"
DEFAULT_MEMCACHED_BACKEND = "default_memcached_backend"
MEMCACHED_IMAGE = "memcached:latest"
MEMCACHED_PORTS = {"11211/tcp": None}
MEMCACHED_ENV: dict = {}
MEMCACHED_CONTAINER_TIMEOUT = 60
MEMCACHED_PREFIX = "cache+memcached://"



"""The pytest-celery plugin provides a set of built-in components called
:ref:`vendors`.

This module is part of the Localstack Broker vendor.
"""

CELERY_LOCALSTACK_BROKER = "celery_localstack_broker"
DEFAULT_LOCALSTACK_BROKER = "default_localstack_broker"
LOCALSTACK_IMAGE = "localstack/localstack"
LOCALSTACK_PORTS = {"4566/tcp": None}
LOCALSTACK_CREDS: dict = {
    "AWS_DEFAULT_REGION": "us-east-1",
    "AWS_ACCESS_KEY_ID": "test",
    "AWS_SECRET_ACCESS_KEY": "test",
}
LOCALSTACK_ENV: dict = {
    **LOCALSTACK_CREDS,
}
LOCALSTACK_CONTAINER_TIMEOUT = 60
LOCALSTACK_PREFIX = "sqs://"



"""The pytest-celery plugin provides a set of built-in components called
:ref:`vendors`.

This module is part of the Redis Broker vendor.
"""

# mypy: disable-error-code="misc"

from __future__ import annotations

import pytest
from pytest_docker_tools import container
from pytest_docker_tools import fxtr

from pytest_celery.vendors.redis.broker.api import RedisTestBroker
from pytest_celery.vendors.redis.container import RedisContainer
from pytest_celery.vendors.redis.defaults import REDIS_CONTAINER_TIMEOUT


@pytest.fixture
def celery_redis_broker(default_redis_broker: RedisContainer) -> RedisTestBroker:
    """Creates a RedisTestBroker instance. Responsible for tearing down the
    node.

    Args:
        default_redis_broker (RedisContainer): Instantiated RedisContainer.
    """
    broker = RedisTestBroker(default_redis_broker)
    yield broker
    broker.teardown()


@pytest.fixture
def default_redis_broker_cls() -> type[RedisContainer]:
    """Default Redis broker container class. Override to apply custom
    configuration globally.

    See also: :ref:`vendor-class`.

    Returns:
        type[RedisContainer]: API for managing the vendor's container.
    """
    return RedisContainer


default_redis_broker = container(
    image="{default_redis_broker_image}",
    ports=fxtr("default_redis_broker_ports"),
    environment=fxtr("default_redis_broker_env"),
    network="{default_pytest_celery_network.name}",
    wrapper_class=RedisContainer,
    timeout=REDIS_CONTAINER_TIMEOUT,
    command=fxtr("default_redis_broker_command"),
)


@pytest.fixture
def default_redis_broker_command(default_redis_broker_cls: type[RedisContainer]) -> list[str]:
    """Command to run the container.

    Args:
        default_redis_broker_cls (type[RedisContainer]): See also: :ref:`vendor-class`.

    Returns:
        list[str]: Docker CMD instruction.
    """
    return default_redis_broker_cls.command()


@pytest.fixture
def default_redis_broker_env(default_redis_broker_cls: type[RedisContainer]) -> dict:
    """Environment variables for this vendor.

    Args:
        default_rabbitmq_broker_cls (type[RedisContainer]): See also: :ref:`vendor-class`.

    Returns:
        dict: Items to pass to the container's environment.
    """
    return default_redis_broker_cls.initial_env()


@pytest.fixture
def default_redis_broker_image(default_redis_broker_cls: type[RedisContainer]) -> str:
    """Sets the image name for this vendor.

    Args:
        default_rabbitmq_broker_cls (type[RedisContainer]): See also: :ref:`vendor-class`.

    Returns:
        str: Docker image name.
    """
    return default_redis_broker_cls.image()


@pytest.fixture
def default_redis_broker_ports(default_redis_broker_cls: type[RedisContainer]) -> dict:
    """Port bindings for this vendor.

    Args:
        default_redis_broker_cls (type[RedisContainer]): See also: :ref:`vendor-class`.

    Returns:
        dict: Port bindings.
    """
    return default_redis_broker_cls.ports()



"""The test setup integrates all of the independent test components into a
single entry point for the test.

This module provides the fixtures for the test setup, which will be used
to prepare the environment for testing.

The test setup is automatically configured according to the individual
components of the given architecture.

A test will be parametrized for each combination of supported celery
backends, brokers, and workers in the test setup.
"""

# mypy: disable-error-code="misc"

from __future__ import annotations

import pytest
from celery import Celery

from pytest_celery.api.backend import CeleryBackendCluster
from pytest_celery.api.broker import CeleryBrokerCluster
from pytest_celery.api.setup import CeleryTestSetup
from pytest_celery.api.worker import CeleryWorkerCluster


@pytest.fixture
def celery_setup_cls() -> type[CeleryTestSetup]:  # type: ignore
    """The setup class to use for the test."""
    return CeleryTestSetup


@pytest.fixture
def celery_setup(  # type: ignore
    celery_setup_cls: type[CeleryTestSetup],
    celery_worker_cluster: CeleryWorkerCluster,
    celery_broker_cluster: CeleryBrokerCluster,
    celery_backend_cluster: CeleryBackendCluster,
    celery_setup_app: Celery,
) -> CeleryTestSetup:
    """Prepares a celery setup ready for testing.

    This fixture provides the entry point for a test.
    This fixture loads all components and immediately prepares the environment for testing.

    Example:
        >>> def test_my_celery_setup(celery_setup: CeleryTestSetup):
        ...     assert celery_setup.ready()
        ...     # do some testing
    """
    setup = celery_setup_cls(
        worker_cluster=celery_worker_cluster,
        broker_cluster=celery_broker_cluster,
        backend_cluster=celery_backend_cluster,
        app=celery_setup_app,
    )

    # Shallow ready check for performance reasons
    assert setup.ready(
        ping=False,
        control=False,
        docker=False,
    )
    yield setup
    setup.teardown()


@pytest.fixture
def celery_setup_name(celery_setup_cls: type[CeleryTestSetup]) -> str:  # type: ignore
    """Fixture interface to the API."""
    return celery_setup_cls.name()


@pytest.fixture
def celery_setup_config(
    celery_setup_cls: type[CeleryTestSetup],
    celery_worker_cluster_config: dict,
) -> dict:
    """Fixture interface to the API."""
    return celery_setup_cls.config(
        celery_worker_cluster_config=celery_worker_cluster_config,
    )


@pytest.fixture
def celery_setup_app(
    celery_setup_cls: type[CeleryTestSetup],
    celery_setup_config: dict,
    celery_setup_name: str,
) -> Celery:
    """Fixture interface to the API."""
    return celery_setup_cls.create_setup_app(
        celery_setup_config=celery_setup_config,
        celery_setup_app_name=celery_setup_name,
    )



"""The pytest-celery plugin provides a set of built-in components called
:ref:`vendors`.

This module is part of the :ref:`built-in-worker` vendor.
"""

# mypy: disable-error-code="misc"

from __future__ import annotations

from types import ModuleType

import pytest
from celery import Celery
from pytest_docker_tools import build
from pytest_docker_tools import container
from pytest_docker_tools import fxtr
from pytest_docker_tools import volume

from pytest_celery.api.worker import CeleryTestWorker
from pytest_celery.vendors.worker.container import CeleryWorkerContainer
from pytest_celery.vendors.worker.defaults import DEFAULT_WORKER_CONTAINER_TIMEOUT
from pytest_celery.vendors.worker.defaults import DEFAULT_WORKER_VOLUME
from pytest_celery.vendors.worker.defaults import WORKER_DOCKERFILE_ROOTDIR


@pytest.fixture
def celery_setup_worker(
    default_worker_cls: type[CeleryTestWorker],
    default_worker_container: CeleryWorkerContainer,
    default_worker_app: Celery,
) -> CeleryTestWorker:
    """Creates a CeleryTestWorker instance. Responsible for tearing down the
    node.

    Args:
        default_worker_cls (type[CeleryTestWorker]): Interface class.
        default_worker_container (CeleryWorkerContainer): Instantiated CeleryWorkerContainer.
        default_worker_app (Celery): Celery app instance.
    """
    worker = default_worker_cls(
        container=default_worker_container,
        app=default_worker_app,
    )
    yield worker
    worker.teardown()


@pytest.fixture
def default_worker_cls() -> type[CeleryTestWorker]:
    """Default worker class. Override to apply custom configuration globally.

    See also: :ref:`vendor-class`.

    Returns:
        type[CeleryTestWorker]: API for managing the vendor's node.
    """
    return CeleryTestWorker


@pytest.fixture
def default_worker_container_cls() -> type[CeleryWorkerContainer]:
    """Default worker container class. Override to apply custom configuration
    globally.

    See also: :ref:`vendor-class`.

    Returns:
        type[CeleryWorkerContainer]: API for managing the vendor's container.
    """
    return CeleryWorkerContainer


@pytest.fixture(scope="session")
def default_worker_container_session_cls() -> type[CeleryWorkerContainer]:
    """Default worker container session class. Override to apply custom
    configuration globally.

    See also: :ref:`vendor-class`.

    Returns:
        type[CeleryWorkerContainer]: API for managing the vendor's container.
    """
    return CeleryWorkerContainer


default_worker_container = container(
    image="{celery_base_worker_image.id}",
    ports=fxtr("default_worker_ports"),
    environment=fxtr("default_worker_env"),
    network="{default_pytest_celery_network.name}",
    volumes={"{default_worker_volume.name}": DEFAULT_WORKER_VOLUME},
    wrapper_class=CeleryWorkerContainer,
    timeout=DEFAULT_WORKER_CONTAINER_TIMEOUT,
    command=fxtr("default_worker_command"),
)

celery_base_worker_image = build(
    path=WORKER_DOCKERFILE_ROOTDIR,
    tag="pytest-celery/components/worker:default",
    buildargs={
        "CELERY_VERSION": fxtr("default_worker_celery_version"),
        "CELERY_LOG_LEVEL": fxtr("default_worker_celery_log_level"),
        "CELERY_WORKER_NAME": fxtr("default_worker_celery_worker_name"),
        "CELERY_WORKER_QUEUE": fxtr("default_worker_celery_worker_queue"),
    },
)

default_worker_volume = volume(
    initial_content=fxtr("default_worker_initial_content"),
)


@pytest.fixture(scope="session")
def default_worker_celery_version(default_worker_container_session_cls: type[CeleryWorkerContainer]) -> str:
    """Celery version for this worker.

    Args:
        default_worker_container_session_cls (type[CeleryWorkerContainer]): See also: :ref:`vendor-class`.

    Returns:
        str: Celery version.
    """
    return default_worker_container_session_cls.version()


@pytest.fixture(scope="session")
def default_worker_celery_log_level(default_worker_container_session_cls: type[CeleryWorkerContainer]) -> str:
    """Log level for this worker.

    Args:
        default_worker_container_session_cls (type[CeleryWorkerContainer]): See also: :ref:`vendor-class`.

    Returns:
        str: Log level.
    """
    return default_worker_container_session_cls.log_level()


@pytest.fixture(scope="session")
def default_worker_celery_worker_name(default_worker_container_session_cls: type[CeleryWorkerContainer]) -> str:
    """Name of the worker.

    Args:
        default_worker_container_session_cls (type[CeleryWorkerContainer]): See also: :ref:`vendor-class`.

    Returns:
        str: Worker name.
    """
    return default_worker_container_session_cls.worker_name()


@pytest.fixture(scope="session")
def default_worker_celery_worker_queue(default_worker_container_session_cls: type[CeleryWorkerContainer]) -> str:
    """Worker queue for this worker.

    Args:
        default_worker_container_session_cls (type[CeleryWorkerContainer]): See also: :ref:`vendor-class`.

    Returns:
        str: Worker queue.
    """
    return default_worker_container_session_cls.worker_queue()


@pytest.fixture
def default_worker_command(default_worker_container_cls: type[CeleryWorkerContainer]) -> list[str]:
    """Command to run the container.

    Args:
        default_worker_container_cls (type[CeleryWorkerContainer]): See also: :ref:`vendor-class`.

    Returns:
        list[str]: Docker CMD instruction.
    """
    return default_worker_container_cls.command()


@pytest.fixture
def default_worker_env(
    default_worker_container_cls: type[CeleryWorkerContainer],
    celery_worker_cluster_config: dict,
) -> dict:
    """Environment variables for this worker.

    Args:
        default_worker_container_cls (type[CeleryWorkerContainer]): See also: :ref:`vendor-class`.
        celery_worker_cluster_config (dict): Broker & Backend clusters configuration.

    Returns:
        dict: Items to pass to the container's environment.
    """
    return default_worker_container_cls.initial_env(celery_worker_cluster_config)


@pytest.fixture
def default_worker_initial_content(
    default_worker_container_cls: type[CeleryWorkerContainer],
    default_worker_app_module: ModuleType,
    default_worker_utils_module: ModuleType,
    default_worker_tasks: set,
    default_worker_signals: set,
    default_worker_app: Celery,
) -> dict:
    """Initial content for this worker's volume.

    This is applied on a worker container when using the following volume configuration:

    .. code-block:: python

        default_worker_container = container(
            ...
            volumes={"{default_worker_volume.name}": DEFAULT_WORKER_VOLUME},
            ...
        )

    .. note::

        More volumes may be added additionally.

    Args:
        default_worker_container_cls (type[CeleryWorkerContainer]): See also: :ref:`vendor-class`.
        default_worker_app_module (ModuleType): App module to inject.
        default_worker_utils_module (ModuleType): Utils module to inject.
        default_worker_tasks (set): Tasks modules to inject.
        default_worker_signals (set): Signals modules to inject.
        default_worker_app (Celery): Celery app to initialize the worker with.

    Returns:
        dict: Initial volume content (dict of files).
    """
    return default_worker_container_cls.initial_content(
        app_module=default_worker_app_module,
        utils_module=default_worker_utils_module,
        worker_tasks=default_worker_tasks,
        worker_signals=default_worker_signals,
        worker_app=default_worker_app,
    )


@pytest.fixture
def default_worker_ports(default_worker_container_cls: type[CeleryWorkerContainer]) -> dict | None:
    """Port bindings for this vendor.

    Args:
        default_worker_container_cls (type[CeleryWorkerContainer]): See also: :ref:`vendor-class`.

    Returns:
        dict: Port bindings.
    """
    return default_worker_container_cls.ports()


@pytest.fixture
def default_worker_app_module(default_worker_container_cls: type[CeleryWorkerContainer]) -> ModuleType:
    """App module for this worker.

    Args:
        default_worker_container_cls (type[CeleryWorkerContainer]): See also: :ref:`vendor-class`.

    Returns:
        ModuleType: App module.
    """
    return default_worker_container_cls.app_module()


@pytest.fixture
def default_worker_utils_module(default_worker_container_cls: type[CeleryWorkerContainer]) -> ModuleType:
    """Utils module for this worker.

    Args:
        default_worker_container_cls (type[CeleryWorkerContainer]): See also: :ref:`vendor-class`.

    Returns:
        ModuleType: Utils module.
    """
    return default_worker_container_cls.utils_module()


@pytest.fixture
def default_worker_tasks(default_worker_container_cls: type[CeleryWorkerContainer]) -> set:
    """Tasks modules set for this worker.

    Args:
        default_worker_container_cls (type[CeleryWorkerContainer]): See also: :ref:`vendor-class`.

    Returns:
        set: Tasks modules.
    """
    return default_worker_container_cls.tasks_modules()


@pytest.fixture
def default_worker_signals(default_worker_container_cls: type[CeleryWorkerContainer]) -> set:
    """Signals modules set for this worker.

    Args:
        default_worker_container_cls (type[CeleryWorkerContainer]): See also: :ref:`vendor-class`.

    Returns:
        set: Signals modules.
    """
    return default_worker_container_cls.signals_modules()


@pytest.fixture
def default_worker_app(celery_setup_app: Celery) -> Celery:
    """Celery app instance for this worker.

    Args:
        celery_setup_app (Celery): See also: :ref:`vendor-class`.

    Returns:
        Celery: Celery app instance.
    """
    return celery_setup_app



"""The pytest-celery plugin provides a set of built-in components called
:ref:`vendors`.

This module is part of the Redis vendor.
"""

from __future__ import annotations

from redis import StrictRedis as Redis

from pytest_celery.api.container import CeleryTestContainer
from pytest_celery.vendors.redis.defaults import REDIS_ENV
from pytest_celery.vendors.redis.defaults import REDIS_IMAGE
from pytest_celery.vendors.redis.defaults import REDIS_PORTS
from pytest_celery.vendors.redis.defaults import REDIS_PREFIX


class RedisContainer(CeleryTestContainer):
    """This class manages the lifecycle of a Redis container."""

    @property
    def client(self) -> Redis | None:
        client = Redis.from_url(
            self.celeryconfig["host_url"],
            decode_responses=True,
        )
        return client

    @property
    def celeryconfig(self) -> dict:
        return {
            "url": self.url,
            "host_url": self.host_url,
            "hostname": self.hostname,
            "port": self.port,
            "vhost": self.vhost,
        }

    @classmethod
    def command(
        cls,
        *args: str,
        debugpy: bool = False,
        wait_for_client: bool = True,
        **kwargs: dict,
    ) -> list[str]:
        return [
            "redis-server",
            "--save",
            "",
            "--appendonly",
            "no",
            "--maxmemory-policy",
            "noeviction",
            "--protected-mode",
            "no",
            *args,
        ]

    @property
    def url(self) -> str:
        return f"{self.prefix()}{self.hostname}/{self.vhost}"

    @property
    def host_url(self) -> str:
        return f"{self.prefix()}localhost:{self.port}/{self.vhost}"

    @property
    def hostname(self) -> str:
        return self.attrs["Config"]["Hostname"]

    @property
    def port(self) -> int:
        return self._wait_port("6379/tcp")

    @property
    def vhost(self) -> str:
        return "0"

    @classmethod
    def version(cls) -> str:
        return cls.image().split(":")[-1]

    @classmethod
    def initial_env(cls) -> dict:
        return REDIS_ENV

    @classmethod
    def image(cls) -> str:
        return REDIS_IMAGE

    @classmethod
    def ports(cls) -> dict:
        return REDIS_PORTS

    @classmethod
    def prefix(cls) -> str:
        return REDIS_PREFIX

    @property
    def ready_prompt(self) -> str | None:
        return "Ready to accept connections"



"""The :ref:`built-in-worker` is added to the test matrix using the fixtures of
this module.

These fixtures will configure the test setup for the built-in Celery
worker by default.

You may override these fixtures to customize the test setup for your
specific needs.
"""

# mypy: disable-error-code="misc"

from __future__ import annotations

import pytest

from pytest_celery.api.worker import CeleryTestWorker
from pytest_celery.api.worker import CeleryWorkerCluster
from pytest_celery.defaults import ALL_CELERY_WORKERS


@pytest.fixture(params=ALL_CELERY_WORKERS)
def celery_worker(request: pytest.FixtureRequest) -> CeleryTestWorker:  # type: ignore
    """Parameterized fixture for all supported celery workers. Responsible for
    tearing down the node.

    This fixture will add parametrization to the test function, so that
    the test will be executed for each supported celery worker.
    """

    worker: CeleryTestWorker = request.getfixturevalue(request.param)
    yield worker
    worker.teardown()


@pytest.fixture
def celery_worker_cluster(celery_worker: CeleryTestWorker) -> CeleryWorkerCluster:  # type: ignore
    """Defines the cluster of worker nodes for the test. Responsible for
    tearing down the cluster.

    To disable the cluster, override this fixture and return None.

    Args:
        celery_worker (CeleryTestWorker): Parameterized fixture for all supported celery workers.

    Returns:
        CeleryWorkerCluster: Single node cluster for all supported celery workers.
    """
    cluster = CeleryWorkerCluster(celery_worker)  # type: ignore
    yield cluster
    cluster.teardown()


@pytest.fixture
def celery_worker_cluster_config(celery_broker_cluster_config: dict, celery_backend_cluster_config: dict) -> dict:
    """Combine the broker and backend cluster configurations.

    Additional configuration can be added.
    """
    return {
        "celery_broker_cluster_config": celery_broker_cluster_config,
        "celery_backend_cluster_config": celery_backend_cluster_config,
    }



"""The ``pytest_docker_tools`` package powers the Docker container management
for the plugin.

This module extends the ``Container`` class from the package to provide
the base API for creating new test containers in a Celery test environment.
"""

from __future__ import annotations

from typing import Any

import pytest_docker_tools
from pytest_docker_tools import wrappers
from pytest_docker_tools.wrappers.container import wait_for_callable
from tenacity import retry
from tenacity import retry_if_exception_type
from tenacity import stop_after_attempt
from tenacity import wait_fixed


class CeleryTestContainer(wrappers.Container):
    """This is an extension of pytest_docker_tools.wrappers.Container, adding
    improved control over the container lifecycle.

    Responsibility Scope:
        Provide useful methods for managing a container instance.
    """

    @property
    def client(self) -> Any:
        """Provides an API client for interacting with the container, if
        available.

        Subclasses should implement this to return an instance of the client
        specific to the service running in the container.

        Raises:
            NotImplementedError: There is not client available by default.

        Returns:
            Any: Client instance.
        """
        raise NotImplementedError("CeleryTestContainer.client")

    @property
    def celeryconfig(self) -> dict:
        """Each container is responsible for providing the configuration values
        required for Celery. This property should be implemented to return the
        configuration values for the specific container.

        Raises:
            NotImplementedError: There is no config available by default.

        Returns:
            dict: Configuration values required for Celery.
        """
        raise NotImplementedError("CeleryTestContainer.celeryconfig")

    @classmethod
    def command(
        cls,
        *args: str,
        debugpy: bool = False,
        wait_for_client: bool = True,
        **kwargs: dict,
    ) -> list[str]:
        """Override the CMD instruction in the Dockerfile.

        This method should be overridden in derived classes to provide the
        specific command and its arguments required to start the container.

        Args:
            *args (str): Additional command-line arguments.
            debugpy (bool): Enable debugpy. Defaults to False.
            wait_for_client (bool): Wait for a debugger to be attached. Defaults to True.
            **kwargs (dict): Additional keyword arguments.

        Raises:
            NotImplementedError: Rely on the Dockerfile if not set otherwise by default.

        Returns:
            list[str]: A list containing the command to run in the container as
                the first element, followed by the command-line arguments.
        """

        raise NotImplementedError("CeleryTestContainer.command")

    def teardown(self) -> None:
        """Teardown the container instance."""

    @property
    def ready_prompt(self) -> str | None:
        """A log string that indicates the container has finished starting up
        and is ready to use.

        Returns:
            str | None: A string to wait for or None for no wait. Defaults to None.
        """
        return None

    @retry(
        stop=stop_after_attempt(3),
        wait=wait_fixed(3),
        retry=retry_if_exception_type(IndexError),
        reraise=True,
    )
    def _wait_port(self, port: str) -> int:
        """Blocks until the specified port is ready.

        Args:
            port (str): Port to wait for.

        Raises:
            ValueError: Port cannot be empty.

        Returns:
            int: Port number.
        """
        if not port:
            raise ValueError("Port cannot be empty")

        while not super().ready():
            pass
        _, p = self.get_addr(port)
        return p

    @retry(
        stop=stop_after_attempt(3),
        wait=wait_fixed(10),
        retry=retry_if_exception_type(pytest_docker_tools.exceptions.TimeoutError),
        reraise=True,
    )
    def _wait_ready(self, timeout: int = 30) -> bool:
        """Wait for the container to be ready by polling the logs for the
        readiness prompt. If no prompt is set, the container is considered
        ready as soon as its live logs are accessible.

        Args:
            timeout (int, optional): Timeout in seconds. Defaults to 30.

        Returns:
            bool: True if ready, False otherwise.
        """
        if self.ready_prompt is not None:
            if self.ready_prompt not in self.logs():
                wait_for_callable(
                    f"Waiting for {self.__class__.__name__}::{self.name} to get ready",
                    lambda: self.ready_prompt in self.logs(),
                    timeout=timeout,
                )

        wait_for_callable(
            f"{self.__class__.__name__}::{self.name} is ready",
            lambda: self.ready_prompt or "" in self.logs(),
        )
        return True

    def ready(self) -> bool:
        """Override the default ready() method to wait for the container to be
        using our waiting logic on top of the default implementation.

        When this method returns False, other attempts will be made until the container
        is ready or stop if other conditions do not allow for it to be ready.
        The underlying implementation of pytest_docker_tools is currently responsible for error raising.

        Returns:
            bool: True if ready, False otherwise.
        """
        if super().ready():
            return self._wait_ready()
        else:
            return False



"""The pytest-celery plugin provides a set of built-in components called
:ref:`vendors`.

This module is part of the RabbitMQ Broker vendor.
"""

# mypy: disable-error-code="misc"

from __future__ import annotations

import pytest
from pytest_docker_tools import container
from pytest_docker_tools import fxtr

from pytest_celery.vendors.rabbitmq.api import RabbitMQTestBroker
from pytest_celery.vendors.rabbitmq.container import RabbitMQContainer
from pytest_celery.vendors.rabbitmq.defaults import RABBITMQ_CONTAINER_TIMEOUT


@pytest.fixture
def celery_rabbitmq_broker(default_rabbitmq_broker: RabbitMQContainer) -> RabbitMQTestBroker:
    """Creates a RabbitMQTestBroker instance. Responsible for tearing down the
    node.

    Args:
        default_rabbitmq_broker (RabbitMQContainer): Instantiated RabbitMQContainer.
    """
    broker = RabbitMQTestBroker(default_rabbitmq_broker)
    yield broker
    broker.teardown()


@pytest.fixture
def default_rabbitmq_broker_cls() -> type[RabbitMQContainer]:
    """Default RabbitMQ broker container class. Override to apply custom
    configuration globally.

    See also: :ref:`vendor-class`.

    Returns:
        type[RabbitMQContainer]: API for managing the vendor's container.
    """
    return RabbitMQContainer


default_rabbitmq_broker = container(
    image="{default_rabbitmq_broker_image}",
    ports=fxtr("default_rabbitmq_broker_ports"),
    environment=fxtr("default_rabbitmq_broker_env"),
    network="{default_pytest_celery_network.name}",
    wrapper_class=RabbitMQContainer,
    timeout=RABBITMQ_CONTAINER_TIMEOUT,
)


@pytest.fixture
def default_rabbitmq_broker_env(default_rabbitmq_broker_cls: type[RabbitMQContainer]) -> dict:
    """Environment variables for this vendor.

    Args:
        default_rabbitmq_broker_cls (type[RabbitMQContainer]): See also: :ref:`vendor-class`.

    Returns:
        dict: Items to pass to the container's environment.
    """
    return default_rabbitmq_broker_cls.initial_env()


@pytest.fixture
def default_rabbitmq_broker_image(default_rabbitmq_broker_cls: type[RabbitMQContainer]) -> str:
    """Sets the image name for this vendor.

    Args:
        default_rabbitmq_broker_cls (type[RabbitMQContainer]): See also: :ref:`vendor-class`.

    Returns:
        str: Docker image name.
    """
    return default_rabbitmq_broker_cls.image()


@pytest.fixture
def default_rabbitmq_broker_ports(default_rabbitmq_broker_cls: type[RabbitMQContainer]) -> dict:
    """Port bindings for this vendor.

    Args:
        default_rabbitmq_broker_cls (type[RabbitMQContainer]): See also: :ref:`vendor-class`.

    Returns:
        dict: Port bindings.
    """
    return default_rabbitmq_broker_cls.ports()



"""Worker components represents Celery's worker instances.

This module provides the base API for creating new worker components by
defining the base classes for worker nodes and clusters.
"""

from __future__ import annotations

import json

from celery import Celery

from pytest_celery.api.base import CeleryTestCluster
from pytest_celery.api.base import CeleryTestNode
from pytest_celery.api.container import CeleryTestContainer
from pytest_celery.vendors.worker.container import CeleryWorkerContainer


class CeleryTestWorker(CeleryTestNode):
    """This is specialized node type for handling celery worker nodes. It is
    used to encapsulate a worker instance.

    Responsibility Scope:
        Managing a celery worker.
    """

    def __init__(self, container: CeleryTestContainer, app: Celery):
        """A celery worker node must be initialized with a celery app.

        Args:
            container (CeleryTestContainer): Container to use for the node.
            app (Celery, optional): Celery app to be accessed from the tests.
        """
        super().__init__(container, app)

        # Helps with autocomplete in the IDE
        self.container: CeleryWorkerContainer

    @property
    def version(self) -> str:
        """Celery version of this worker node."""
        return self.container.version()

    @property
    def log_level(self) -> str:
        """Celery log level of this worker node."""
        return self.container.log_level()

    @property
    def worker_name(self) -> str:
        """Celery test worker node name."""
        return self.container.worker_name()

    @property
    def worker_queue(self) -> str:
        """Celery queue for this worker node."""
        return self.container.worker_queue()

    def hostname(self) -> str:
        """Hostname of the worker node."""
        return f"{self.worker_name}@{super().hostname()}"

    def get_running_processes_info(
        self,
        columns: list[str] | None = None,
        filters: dict[str, str] | None = None,
    ) -> list[dict]:
        """Get running processes info on the container of this node.

        Possible columns:
            - pid
            - name
            - username
            - cmdline
            - cpu_percent
            - memory_percent
            - create_time

        Args:
            columns (list[str] | None, optional): Columns to query. Defaults to None (all).
            filters (dict[str, str] | None, optional): Filters to apply. Defaults to None.

        Raises:
            RuntimeError: If the command fails.

        Returns:
            list[dict]: List of processes info per requested columns.
        """
        # Use special vendors/worker/content/utils.py module
        exit_code, output = self.container.exec_run(
            f'python -c "from utils import get_running_processes_info; print(get_running_processes_info({columns!r}))"'
        )

        if exit_code != 0:
            raise RuntimeError(f"Failed to get processes info: {output}")

        decoded_str = output.decode("utf-8")
        output = json.loads(decoded_str)

        if filters:
            output = [item for item in output if all(item.get(key) == value for key, value in filters.items())]

        return output


class CeleryWorkerCluster(CeleryTestCluster):
    """This is a specialized cluster type for handling celery workers. It is
    used to define which worker instances are available for the test.

    Responsibility Scope:
        Provude useful methods for managing a cluster of celery workers.
    """

    @property
    def versions(self) -> set[str]:
        """Celery versions of all workers in this cluster."""
        return {worker.version for worker in self}  # type: ignore



"""The pytest-celery plugin provides a set of built-in components called
:ref:`vendors`.

This module is part of the Redis Backend vendor.
"""

CELERY_REDIS_BACKEND = "celery_redis_backend"
DEFAULT_REDIS_BACKEND = "default_redis_backend"



"""Default plugin configurations.

This module contains the default configurations for the pytest-celery
plugin. It also contains the automatic collections that are used to
parametrize the user's tests according to the components they need to
run against. By default, all possible combinations are used.
"""

# flake8: noqa

from __future__ import annotations

from pytest_docker_tools import network

from pytest_celery.vendors import _is_vendor_installed
from pytest_celery.vendors.localstack.defaults import CELERY_LOCALSTACK_BROKER
from pytest_celery.vendors.localstack.defaults import *
from pytest_celery.vendors.memcached.defaults import CELERY_MEMCACHED_BACKEND
from pytest_celery.vendors.memcached.defaults import *
from pytest_celery.vendors.rabbitmq.defaults import CELERY_RABBITMQ_BROKER
from pytest_celery.vendors.rabbitmq.defaults import *
from pytest_celery.vendors.redis.backend.defaults import CELERY_REDIS_BACKEND
from pytest_celery.vendors.redis.backend.defaults import *
from pytest_celery.vendors.redis.broker.defaults import CELERY_REDIS_BROKER
from pytest_celery.vendors.redis.broker.defaults import *
from pytest_celery.vendors.redis.defaults import *
from pytest_celery.vendors.worker.defaults import CELERY_SETUP_WORKER
from pytest_celery.vendors.worker.defaults import *

####################################################################################
# Automatic components
####################################################################################

# The following collections are used to parametrize the user's tests
# according to the components they need to run against. By default,
# all possible combinations are used.

# When a new component is added to the corresponding collection it
# will automatically add it to the parametrization of every (relevant) test IMPLICITLY!
# Tests that do not rely on default parametrization will not be affected.


ALL_CELERY_BACKENDS = set()
ALL_CELERY_BROKERS = set()

if _is_vendor_installed("redis"):
    ALL_CELERY_BACKENDS.add(CELERY_REDIS_BACKEND)
    ALL_CELERY_BROKERS.add(CELERY_REDIS_BROKER)

if _is_vendor_installed("rabbitmq"):
    # Uses Kombu
    ALL_CELERY_BROKERS.add(CELERY_RABBITMQ_BROKER)

# Memcached is disabled by default regardless of its availability due to its experimental status.
if _is_vendor_installed("memcached") and False:
    ALL_CELERY_BACKENDS.add(CELERY_MEMCACHED_BACKEND)

# Localstack is disabled by default regardless of its availability due to its beta status.
if _is_vendor_installed("localstack") and False:
    # Uses Kombu
    ALL_CELERY_BROKERS.add(CELERY_LOCALSTACK_BROKER)

# Worker setup is assumed to be always available.
ALL_CELERY_WORKERS = (CELERY_SETUP_WORKER,)

####################################################################################
# Fixtures
####################################################################################

# These fixtures are used by pytest-celery plugin to setup the environment for the user's tests.
# They function as the main interface into the test environment's components.
# They are directly affected by the automatic collections: ALL_CELERY_WORKERS, ALL_CELERY_BACKENDS, ALL_CELERY_BROKERS.

CELERY_SETUP = "celery_setup"
CELERY_WORKER = "celery_worker"
CELERY_WORKER_CLUSTER = "celery_worker_cluster"
CELERY_BACKEND = "celery_backend"
CELERY_BACKEND_CLUSTER = "celery_backend_cluster"
CELERY_BROKER = "celery_broker"
CELERY_BROKER_CLUSTER = "celery_broker_cluster"

####################################################################################
# Docker
####################################################################################

# Default timeouts for Docker-related operations in seconds.

# Defines how long to wait for a Docker container to get ready.
CONTAINER_TIMEOUT = 60

# Specifies the maximum duration to wait for a task result.
# It's primarily used when testing Celery tasks to ensure they complete within a reasonable timeframe.
# and is set for a recommended value for most use cases.
RESULT_TIMEOUT = 60

default_pytest_celery_network = network()



"""Official pytest plugin for Celery."""

# flake8: noqa


__version__ = "v1.2.1"
__author__ = "Tomer Nosrati"
__contact__ = "tomer.nosrati@gmail.com"
__homepage__ = "https://pytest-celery.readthedocs.io/"
__docformat__ = "restructuredtext"


import re
from collections import namedtuple

from pytest_celery.api.backend import CeleryBackendCluster
from pytest_celery.api.backend import CeleryTestBackend
from pytest_celery.api.base import CeleryTestCluster
from pytest_celery.api.base import CeleryTestNode
from pytest_celery.api.broker import CeleryBrokerCluster
from pytest_celery.api.broker import CeleryTestBroker
from pytest_celery.api.container import CeleryTestContainer
from pytest_celery.api.setup import CeleryTestSetup
from pytest_celery.api.worker import CeleryTestWorker
from pytest_celery.api.worker import CeleryWorkerCluster
from pytest_celery.defaults import *
from pytest_celery.fixtures.backend import celery_backend
from pytest_celery.fixtures.backend import celery_backend_cluster
from pytest_celery.fixtures.backend import celery_backend_cluster_config
from pytest_celery.fixtures.broker import celery_broker
from pytest_celery.fixtures.broker import celery_broker_cluster
from pytest_celery.fixtures.broker import celery_broker_cluster_config
from pytest_celery.fixtures.setup import celery_setup
from pytest_celery.fixtures.setup import celery_setup_app
from pytest_celery.fixtures.setup import celery_setup_cls
from pytest_celery.fixtures.setup import celery_setup_config
from pytest_celery.fixtures.setup import celery_setup_name
from pytest_celery.fixtures.worker import celery_worker
from pytest_celery.fixtures.worker import celery_worker_cluster
from pytest_celery.fixtures.worker import celery_worker_cluster_config
from pytest_celery.vendors import _is_vendor_installed

if _is_vendor_installed("localstack"):
    from pytest_celery.vendors.localstack.api import LocalstackTestBroker
    from pytest_celery.vendors.localstack.container import LocalstackContainer
    from pytest_celery.vendors.localstack.defaults import *
    from pytest_celery.vendors.localstack.fixtures import celery_localstack_broker
    from pytest_celery.vendors.localstack.fixtures import default_localstack_broker
    from pytest_celery.vendors.localstack.fixtures import default_localstack_broker_cls
    from pytest_celery.vendors.localstack.fixtures import default_localstack_broker_env
    from pytest_celery.vendors.localstack.fixtures import default_localstack_broker_image
    from pytest_celery.vendors.localstack.fixtures import default_localstack_broker_ports

if _is_vendor_installed("memcached"):
    from pytest_celery.vendors.memcached.api import MemcachedTestBackend
    from pytest_celery.vendors.memcached.container import MemcachedContainer
    from pytest_celery.vendors.memcached.defaults import *
    from pytest_celery.vendors.memcached.fixtures import celery_memcached_backend
    from pytest_celery.vendors.memcached.fixtures import default_memcached_backend
    from pytest_celery.vendors.memcached.fixtures import default_memcached_backend_cls
    from pytest_celery.vendors.memcached.fixtures import default_memcached_backend_env
    from pytest_celery.vendors.memcached.fixtures import default_memcached_backend_image
    from pytest_celery.vendors.memcached.fixtures import default_memcached_backend_ports

if _is_vendor_installed("rabbitmq"):
    from pytest_celery.vendors.rabbitmq.api import RabbitMQTestBroker
    from pytest_celery.vendors.rabbitmq.container import RabbitMQContainer
    from pytest_celery.vendors.rabbitmq.defaults import *
    from pytest_celery.vendors.rabbitmq.fixtures import celery_rabbitmq_broker
    from pytest_celery.vendors.rabbitmq.fixtures import default_rabbitmq_broker
    from pytest_celery.vendors.rabbitmq.fixtures import default_rabbitmq_broker_cls
    from pytest_celery.vendors.rabbitmq.fixtures import default_rabbitmq_broker_env
    from pytest_celery.vendors.rabbitmq.fixtures import default_rabbitmq_broker_image
    from pytest_celery.vendors.rabbitmq.fixtures import default_rabbitmq_broker_ports

if _is_vendor_installed("redis"):
    from pytest_celery.vendors.redis.backend.api import RedisTestBackend
    from pytest_celery.vendors.redis.backend.defaults import *
    from pytest_celery.vendors.redis.backend.fixtures import celery_redis_backend
    from pytest_celery.vendors.redis.backend.fixtures import default_redis_backend
    from pytest_celery.vendors.redis.backend.fixtures import default_redis_backend_cls
    from pytest_celery.vendors.redis.backend.fixtures import default_redis_backend_command
    from pytest_celery.vendors.redis.backend.fixtures import default_redis_backend_env
    from pytest_celery.vendors.redis.backend.fixtures import default_redis_backend_image
    from pytest_celery.vendors.redis.backend.fixtures import default_redis_backend_ports
    from pytest_celery.vendors.redis.broker.api import RedisTestBroker
    from pytest_celery.vendors.redis.broker.defaults import *
    from pytest_celery.vendors.redis.broker.fixtures import celery_redis_broker
    from pytest_celery.vendors.redis.broker.fixtures import default_redis_broker
    from pytest_celery.vendors.redis.broker.fixtures import default_redis_broker_cls
    from pytest_celery.vendors.redis.broker.fixtures import default_redis_broker_command
    from pytest_celery.vendors.redis.broker.fixtures import default_redis_broker_env
    from pytest_celery.vendors.redis.broker.fixtures import default_redis_broker_image
    from pytest_celery.vendors.redis.broker.fixtures import default_redis_broker_ports
    from pytest_celery.vendors.redis.container import RedisContainer
    from pytest_celery.vendors.redis.defaults import *

if _is_vendor_installed("worker"):
    from pytest_celery.vendors.worker.container import CeleryWorkerContainer
    from pytest_celery.vendors.worker.content import app
    from pytest_celery.vendors.worker.content import utils
    from pytest_celery.vendors.worker.fixtures import celery_base_worker_image
    from pytest_celery.vendors.worker.fixtures import celery_setup_worker
    from pytest_celery.vendors.worker.fixtures import default_worker_app
    from pytest_celery.vendors.worker.fixtures import default_worker_app_module
    from pytest_celery.vendors.worker.fixtures import default_worker_celery_log_level
    from pytest_celery.vendors.worker.fixtures import default_worker_celery_version
    from pytest_celery.vendors.worker.fixtures import default_worker_celery_worker_name
    from pytest_celery.vendors.worker.fixtures import default_worker_celery_worker_queue
    from pytest_celery.vendors.worker.fixtures import default_worker_cls
    from pytest_celery.vendors.worker.fixtures import default_worker_command
    from pytest_celery.vendors.worker.fixtures import default_worker_container
    from pytest_celery.vendors.worker.fixtures import default_worker_container_cls
    from pytest_celery.vendors.worker.fixtures import default_worker_container_session_cls
    from pytest_celery.vendors.worker.fixtures import default_worker_env
    from pytest_celery.vendors.worker.fixtures import default_worker_initial_content
    from pytest_celery.vendors.worker.fixtures import default_worker_ports
    from pytest_celery.vendors.worker.fixtures import default_worker_signals
    from pytest_celery.vendors.worker.fixtures import default_worker_tasks
    from pytest_celery.vendors.worker.fixtures import default_worker_utils_module
    from pytest_celery.vendors.worker.fixtures import default_worker_volume
    from pytest_celery.vendors.worker.tasks import *
    from pytest_celery.vendors.worker.volume import WorkerInitialContent


version_info_t = namedtuple(
    "version_info_t",
    (
        "major",
        "minor",
        "micro",
        "releaselevel",
        "serial",
    ),
)


# Required for RTD to build
_temp = re.match(r"v?(\d+)\.(\d+)\.(\d+)([a-zA-Z0-9]+)?", __version__).groups()  # type: ignore
VERSION = version_info = version_info_t(int(_temp[0]), int(_temp[1]), int(_temp[2]), _temp[3] or "", "")
del _temp
del re



"""The pytest-celery plugin provides a set of built-in components called
:ref:`vendors`.

This module is part of the RabbitMQ Broker vendor.
"""

from __future__ import annotations

from kombu import Connection

from pytest_celery.api.container import CeleryTestContainer
from pytest_celery.vendors.rabbitmq.defaults import RABBITMQ_ENV
from pytest_celery.vendors.rabbitmq.defaults import RABBITMQ_IMAGE
from pytest_celery.vendors.rabbitmq.defaults import RABBITMQ_PORTS
from pytest_celery.vendors.rabbitmq.defaults import RABBITMQ_PREFIX


class RabbitMQContainer(CeleryTestContainer):
    """This class manages the lifecycle of a RabbitMQ container."""

    @property
    def client(self) -> Connection:
        client = Connection(
            self.celeryconfig["host_url"],
            port=self.celeryconfig["port"],
        )
        return client

    @property
    def celeryconfig(self) -> dict:
        return {
            "url": self.url,
            "host_url": self.host_url,
            "hostname": self.hostname,
            "port": self.port,
            "vhost": self.vhost,
        }

    @property
    def url(self) -> str:
        return f"{self.prefix()}{self.hostname}/{self.vhost}"

    @property
    def host_url(self) -> str:
        return f"{self.prefix()}localhost:{self.port}/{self.vhost}"

    @property
    def hostname(self) -> str:
        return self.attrs["Config"]["Hostname"]

    @property
    def port(self) -> int:
        return self._wait_port("5672/tcp")

    @property
    def vhost(self) -> str:
        return "/"

    @classmethod
    def version(cls) -> str:
        return cls.image().split(":")[-1]

    @classmethod
    def initial_env(cls) -> dict:
        return RABBITMQ_ENV

    @classmethod
    def image(cls) -> str:
        return RABBITMQ_IMAGE

    @classmethod
    def ports(cls) -> dict:
        return RABBITMQ_PORTS

    @classmethod
    def prefix(cls) -> str:
        return RABBITMQ_PREFIX

    @property
    def ready_prompt(self) -> str | None:
        return "Server startup complete"



"""Broker components represents Celery's broker instances.

This module provides the base API for creating new broker components by
defining the base classes for broker nodes and clusters.
"""

from __future__ import annotations

from pytest_celery.api.base import CeleryTestCluster
from pytest_celery.api.base import CeleryTestNode
from pytest_celery.defaults import DEFAULT_WORKER_ENV


class CeleryTestBroker(CeleryTestNode):
    """This is specialized node type for handling celery brokers nodes. It is
    used to encapsulate a broker instance.

    Responsibility Scope:
        Handling broker specific requirements and configuration.
    """

    @classmethod
    def default_config(cls) -> dict:
        """Default node configurations if not overridden by the user."""
        return {
            "url": DEFAULT_WORKER_ENV["CELERY_BROKER_URL"],
            "host_url": DEFAULT_WORKER_ENV["CELERY_BROKER_URL"],
        }

    def restart(self, reload_container: bool = True, force: bool = False) -> None:
        """Override restart method to update the app broker url with new
        container values."""
        super().restart(reload_container, force)
        if self.app:
            self.app.conf.update(
                broker_url=self.config()["host_url"],
            )


class CeleryBrokerCluster(CeleryTestCluster):
    """This is a specialized cluster type for handling celery brokers. It is
    used to define which broker instances are available for the test.

    Responsibility Scope:
        Provude useful methods for managing a cluster of celery brokers.
    """

    @classmethod
    def default_config(cls) -> dict:
        """Default cluster configurations if not overridden by the user."""
        return {
            "urls": [DEFAULT_WORKER_ENV["CELERY_BROKER_URL"]],
            "host_urls": [DEFAULT_WORKER_ENV["CELERY_BROKER_URL"]],
        }



"""The pytest-celery plugin provides a set of built-in components called
:ref:`vendors`.

This module is part of the Redis vendor.
"""

REDIS_IMAGE = "redis:latest"
REDIS_PORTS = {"6379/tcp": None}
REDIS_ENV: dict = {}
REDIS_CONTAINER_TIMEOUT = 60
REDIS_PREFIX = "redis://"



"""The pytest-celery plugin provides a set of built-in components called
:ref:`vendors`.

This module is part of the Memcached Backend vendor.
"""

from __future__ import annotations

import memcache

from pytest_celery.api.container import CeleryTestContainer
from pytest_celery.vendors.memcached.defaults import MEMCACHED_ENV
from pytest_celery.vendors.memcached.defaults import MEMCACHED_IMAGE
from pytest_celery.vendors.memcached.defaults import MEMCACHED_PORTS
from pytest_celery.vendors.memcached.defaults import MEMCACHED_PREFIX


class MemcachedContainer(CeleryTestContainer):
    """This class manages the lifecycle of a Memcached container."""

    @property
    def client(self) -> memcache.Client:
        conf = self.celeryconfig
        servers = [f"{conf['host_url'][:-1].split('://')[-1]}"]
        client = memcache.Client(servers)
        return client

    @property
    def celeryconfig(self) -> dict:
        return {
            "url": self.url,
            "host_url": self.host_url,
            "hostname": self.hostname,
            "port": self.port,
        }

    @property
    def url(self) -> str:
        return f"{self.prefix()}{self.hostname}/"

    @property
    def host_url(self) -> str:
        return f"{self.prefix()}localhost:{self.port}/"

    @property
    def hostname(self) -> str:
        return self.attrs["Config"]["Hostname"]

    @property
    def port(self) -> int:
        return self._wait_port("11211/tcp")

    @classmethod
    def version(cls) -> str:
        return cls.image().split(":")[-1]

    @classmethod
    def initial_env(cls) -> dict:
        return MEMCACHED_ENV

    @classmethod
    def image(cls) -> str:
        return MEMCACHED_IMAGE

    @classmethod
    def ports(cls) -> dict:
        return MEMCACHED_PORTS

    @classmethod
    def prefix(cls) -> str:
        return MEMCACHED_PREFIX



"""Every backend component is added to the test matrix using the fixtures of
this module.

These fixtures will configure the test setup for all supported celery
backends by default. Every backend will be executed as a separate test
case, and the test will be executed for each supported celery backend.

You may override these fixtures to customize the test setup for your
specific needs.
"""

# mypy: disable-error-code="misc"

from __future__ import annotations

import pytest

from pytest_celery.api.backend import CeleryBackendCluster
from pytest_celery.api.backend import CeleryTestBackend
from pytest_celery.defaults import ALL_CELERY_BACKENDS
from pytest_celery.defaults import CELERY_BACKEND_CLUSTER


@pytest.fixture(params=ALL_CELERY_BACKENDS)
def celery_backend(request: pytest.FixtureRequest) -> CeleryTestBackend:  # type: ignore
    """Parameterized fixture for all supported celery backends. Responsible for
    tearing down the node.

    This fixture will add parametrization to the test function, so that
    the test will be executed for each supported celery backend.
    """
    backend: CeleryTestBackend = request.getfixturevalue(request.param)
    yield backend
    backend.teardown()


@pytest.fixture
def celery_backend_cluster(celery_backend: CeleryTestBackend) -> CeleryBackendCluster:  # type: ignore
    """Defines the cluster of backend nodes for the test. Responsible for
    tearing down the cluster.

    To disable the cluster, override this fixture and return None.

    Args:
        celery_backend (CeleryTestBackend): Parameterized fixture for all supported celery backends.

    Returns:
        CeleryBackendCluster: Single node cluster for all supported celery backends.
    """
    cluster = CeleryBackendCluster(celery_backend)  # type: ignore
    yield cluster
    cluster.teardown()


@pytest.fixture
def celery_backend_cluster_config(request: pytest.FixtureRequest) -> dict | None:
    """Attempts to compile the celery configuration from the cluster."""
    try:
        use_default_config = pytest.fail.Exception
        cluster: CeleryBackendCluster = request.getfixturevalue(CELERY_BACKEND_CLUSTER)
        return cluster.config() if cluster else None
    except use_default_config:
        return CeleryBackendCluster.default_config()



"""The pytest-celery plugin provides a set of built-in components called
:ref:`vendors`.

This module is part of the Redis Backend vendor.
"""

from __future__ import annotations

import gc

from celery.result import AsyncResult

from pytest_celery.api.backend import CeleryTestBackend


class RedisTestBackend(CeleryTestBackend):
    def teardown(self) -> None:
        """When a test that has a AsyncResult object is finished there's a race
        condition between the AsyncResult object and the Redis container.

        The AsyncResult object tries to release the connection but the
        Redis container has already exited.
        """
        # First, force a garbage collection to clean up unreachable objects
        gc.collect()

        # Next, find all live AsyncResult objects and clean them up
        async_results = [obj for obj in gc.get_objects() if isinstance(obj, AsyncResult)]

        for async_result in async_results:
            try:
                # Remove the backend reference to prevent interaction with Redis
                async_result.backend = None
            except Exception:
                pass

        super().teardown()



"""The node/cluster set are designed to provide a common interface for
encapsulating the implementation of test components and their integration with
the test framework.

This module provides the base API for creating new components by
defining the base classes for nodes and clusters.
"""

from __future__ import annotations

from abc import abstractmethod
from typing import Iterator

import pytest_docker_tools
from celery import Celery
from pytest_docker_tools.wrappers.container import wait_for_callable

from pytest_celery.api.container import CeleryTestContainer
from pytest_celery.defaults import CONTAINER_TIMEOUT
from pytest_celery.defaults import RESULT_TIMEOUT


class CeleryTestNode:
    """This is the logical representation of a container instance. It is used
    to provide a common interface for interacting with the container regardless
    of the underlying implementation.

    Responsibility Scope:
        The node's responsibility is to wrap the container and provide
        useful methods for interacting with it.
    """

    def __init__(self, container: CeleryTestContainer, app: Celery = None) -> None:
        """Setup the base components of a CeleryTestNode.

        Args:
            container (CeleryTestContainer): Container to use for the node.
            app (Celery, optional): Celery app. Defaults to None.
        """
        self._container = container
        self._app = app

    @property
    def container(self) -> CeleryTestContainer:
        """Underlying container for the node."""
        return self._container

    @property
    def app(self) -> Celery:
        """Celery app for the node if available."""
        return self._app

    def __eq__(self, other: object) -> bool:
        """Two nodes are equal if they have the same container and Celery
        app."""
        if isinstance(other, CeleryTestNode):
            return all(
                (
                    self.container == other.container,
                    self.app == other.app,
                )
            )
        return False

    @classmethod
    def default_config(cls) -> dict:
        """Default node configurations if not overridden by the user."""
        return {}

    def ready(self) -> bool:
        """Waits until the node is ready or raise an exception if it fails to
        boot up."""
        return self.container.ready()

    def config(self, *args: tuple, **kwargs: dict) -> dict:
        """Compile the configurations required for Celery from this node."""
        config = self.container.celeryconfig

        if not args and not kwargs:
            return config

        for key, value in kwargs.items():
            config[key] = value

            if key == "vhost":
                vhost = str(value)
                config["url"] = f"{config['url'][:-1].rstrip('/')}/{vhost}"
                config["host_url"] = f"{config['host_url'][:-1].rstrip('/')}/{vhost}"
                config[key] = vhost

        return config

    def logs(self) -> str:
        """Get the logs of the underlying container."""
        return self.container.logs()

    def name(self) -> str:
        """Get the name of this node."""
        return self.container.name

    def hostname(self) -> str:
        """Get the hostname of this node."""
        return self.container.id[:12]

    def kill(self, signal: str | int = "SIGKILL", reload_container: bool = True) -> None:
        """Kill the underlying container.

        Args:
            signal (str | int, optional): Signal to send to the container. Defaults to "SIGKILL".
            reload_container (bool, optional): Reload the container object after killing it. Defaults to True.
        """
        if self.container.status == "running":
            self.container.kill(signal=signal)
        if reload_container:
            self.container.reload()

    def restart(self, reload_container: bool = True, force: bool = False) -> None:
        """Restart the underlying container.

        Args:
            reload_container (bool, optional): Reload the container object after restarting it. Defaults to True.
            force (bool, optional): Kill the container before restarting it. Defaults to False.
        """
        if force:
            # Use SIGTERM to allow the container to gracefully shutdown
            self.kill(signal="SIGTERM", reload_container=reload_container)
        self.container.restart(timeout=CONTAINER_TIMEOUT)
        if reload_container:
            self.container.reload()
        if self.app:
            self.app.set_current()

    def teardown(self) -> None:
        """Teardown the node."""
        self.container.teardown()

    def wait_for_log(self, log: str, message: str = "", timeout: int = RESULT_TIMEOUT) -> None:
        """Wait for a log to appear in the container.

        Args:
            log (str): Log to wait for.
            message (str, optional): Message to display while waiting. Defaults to "".
            timeout (int, optional): Timeout in seconds. Defaults to RESULT_TIMEOUT.
        """
        message = message or f"Waiting for worker container '{self.name()}' to log -> {log}"
        wait_for_callable(message=message, func=lambda: log in self.logs(), timeout=timeout)

    def assert_log_exists(self, log: str, message: str = "", timeout: int = RESULT_TIMEOUT) -> None:
        """Assert that a log exists in the container.

        Args:
            log (str): Log to assert.
            message (str, optional): Message to display while waiting. Defaults to "".
            timeout (int, optional): Timeout in seconds. Defaults to RESULT_TIMEOUT.
        """
        try:
            self.wait_for_log(log, message, timeout)
        except pytest_docker_tools.exceptions.TimeoutError:
            assert False, f"Worker container '{self.name()}' did not log -> {log} within {timeout} seconds"

    def assert_log_does_not_exist(self, log: str, message: str = "", timeout: int = 1) -> None:
        """Assert that a log does not exist in the container.

        Args:
            log (str): Log to assert.
            message (str, optional): Message to display while waiting. Defaults to "".
            timeout (int, optional): Timeout in seconds. Defaults to 1.
        """
        message = message or f"Waiting for worker container '{self.name()}' to not log -> {log}"
        try:
            self.wait_for_log(log, message, timeout)
        except pytest_docker_tools.exceptions.TimeoutError:
            return
        assert False, f"Worker container '{self.name()}' logged -> {log} within {timeout} seconds"


class CeleryTestCluster:
    """This is a collection of CeleryTestNodes. It is used to collect the test
    nodes into a single object for easier management.

    Responsibility Scope:
        The cluster's responsibility is to define which nodes will be used for
        the test.
    """

    def __init__(self, *nodes: tuple[CeleryTestNode | CeleryTestContainer]) -> None:
        """Setup the base components of a CeleryTestCluster.

        Args:
            *nodes (tuple[CeleryTestNode | CeleryTestContainer]): Nodes to use for the cluster.

        Raises:
            ValueError: At least one node is required.
            TypeError: All nodes must be CeleryTestNode or CeleryTestContainer
        """
        if not nodes:
            raise ValueError("At least one node is required")
        if len(nodes) == 1 and isinstance(nodes[0], list):
            nodes = tuple(node for node in nodes[0])
        if not all(isinstance(node, (CeleryTestNode, CeleryTestContainer)) for node in nodes):
            raise TypeError("All nodes must be CeleryTestNode or CeleryTestContainer")

        self.nodes = nodes  # type: ignore

    @property
    def nodes(self) -> tuple[CeleryTestNode]:
        """Get the nodes of the cluster."""
        return self._nodes

    @nodes.setter
    def nodes(self, nodes: tuple[CeleryTestNode | CeleryTestContainer]) -> None:
        """Set the nodes of the cluster.

        Args:
            nodes (tuple[CeleryTestNode | CeleryTestContainer]): Nodes to use for the cluster.
        """
        self._nodes = self._set_nodes(*nodes)  # type: ignore

    def __iter__(self) -> Iterator[CeleryTestNode]:
        """Iterate over the nodes of the cluster."""
        return iter(self.nodes)

    def __getitem__(self, index: int) -> CeleryTestNode:
        """Get a node from the cluster by index."""
        return self.nodes[index]

    def __len__(self) -> int:
        """Get the number of nodes in the cluster."""
        return len(self.nodes)

    def __eq__(self, other: object) -> bool:
        """Two clusters are equal if they have the same nodes."""
        if isinstance(other, CeleryTestCluster):
            if len(self) == len(other):
                for node in self:
                    if node not in other:
                        return False
        return False

    @classmethod
    def default_config(cls) -> dict:
        """Default cluster configurations if not overridden by the user."""
        return {}

    @abstractmethod
    def _set_nodes(
        self,
        *nodes: tuple[CeleryTestNode | CeleryTestContainer],
        node_cls: type[CeleryTestNode] = CeleryTestNode,
    ) -> tuple[CeleryTestNode]:
        """Set the nodes of the cluster.

        Args:
            *nodes (tuple[CeleryTestNode | CeleryTestContainer]): Nodes to use for the cluster.
            node_cls (type[CeleryTestNode], optional): Node class to use. Defaults to CeleryTestNode.

        Returns:
            tuple[CeleryTestNode]: Nodes to use for the cluster.
        """
        return tuple(
            (
                node_cls(node)
                if isinstance(
                    node,
                    CeleryTestContainer,
                )
                else node
            )
            for node in nodes
        )  # type: ignore

    def ready(self) -> bool:
        """Waits until the cluster is ready or raise an exception if any of the
        nodes fail to boot up."""
        return all(node.ready() for node in self)

    def config(self, *args: tuple, **kwargs: dict) -> dict:
        """Compile the configurations required for Celery from this cluster."""
        config = [node.config() for node in self]
        return {
            "urls": [c["url"] for c in config],
            "host_urls": [c["host_url"] for c in config],
        }

    def teardown(self) -> None:
        """Teardown the cluster."""
        # Nodes teardown themselves, so we just need to clear the cluster
        # if there is any cleanup to do



"""The pytest-celery plugin provides a set of built-in components called
:ref:`vendors`.

This module is part of the :ref:`built-in-worker` vendor.
"""

from __future__ import annotations

from types import ModuleType

from celery import Celery

from pytest_celery.api.container import CeleryTestContainer
from pytest_celery.vendors.worker.defaults import DEFAULT_WORKER_ENV
from pytest_celery.vendors.worker.defaults import DEFAULT_WORKER_LOG_LEVEL
from pytest_celery.vendors.worker.defaults import DEFAULT_WORKER_NAME
from pytest_celery.vendors.worker.defaults import DEFAULT_WORKER_PORTS
from pytest_celery.vendors.worker.defaults import DEFAULT_WORKER_QUEUE
from pytest_celery.vendors.worker.defaults import DEFAULT_WORKER_VERSION
from pytest_celery.vendors.worker.volume import WorkerInitialContent


class CeleryWorkerContainer(CeleryTestContainer):
    """This is the base class for all Celery worker containers. It is
    preconfigured for a built-in Celery worker image and should be customized
    for your own worker image.

    The purpose of this class is manipulating the container volume and
    configurations to warm up the worker container according to the test case requirements.

    Responsibility Scope:
        Prepare the worker container with the required filesystem, configurations and
        dependencies of your project.
    """

    @classmethod
    def command(
        cls,
        *args: str,
        debugpy: bool = False,
        wait_for_client: bool = True,
        **kwargs: dict,
    ) -> list[str]:
        args = args or tuple()
        cmd = list()

        if debugpy:
            cmd.extend(
                [
                    "python",
                    "-m",
                    "debugpy",
                    "--listen",
                    "0.0.0.0:5678",
                ]
            )

            if wait_for_client:
                cmd.append("--wait-for-client")

            cmd.append("-m")

        cmd.extend(
            [
                "celery",
                "-A",
                "app",
                "worker",
                f"--loglevel={cls.log_level()}",
                "-n",
                f"{cls.worker_name()}@%h",
                "-Q",
                f"{cls.worker_queue()}",
                *args,
            ]
        )

        return cmd

    def _wait_port(self, port: str) -> int:
        # Not needed for worker container
        raise NotImplementedError

    @property
    def ready_prompt(self) -> str:
        return "ready."

    @classmethod
    def version(cls) -> str:
        """Celery version to use for the worker container."""
        return DEFAULT_WORKER_VERSION

    @classmethod
    def log_level(cls) -> str:
        """Celery worker log level."""
        return DEFAULT_WORKER_LOG_LEVEL

    @classmethod
    def worker_name(cls) -> str:
        """Celery worker name."""
        return DEFAULT_WORKER_NAME

    @classmethod
    def worker_queue(cls) -> str:
        """Celery worker queue."""
        return DEFAULT_WORKER_QUEUE

    @classmethod
    def app_module(cls) -> ModuleType:
        """A preconfigured module that contains the Celery app instance.

        The module is manipulated at runtime to inject the required
        configurations from the test case.
        """
        from pytest_celery.vendors.worker.content import app

        return app

    @classmethod
    def utils_module(cls) -> ModuleType:
        """A utility helper module for running python code in the worker
        container context."""
        from pytest_celery.vendors.worker.content import utils

        return utils

    @classmethod
    def tasks_modules(cls) -> set:
        """Tasks modules."""
        from pytest_celery.vendors.worker import tasks as default_tasks

        return {default_tasks}

    @classmethod
    def signals_modules(cls) -> set:
        """Signals handlers modules.

        This is an optional feature that can be used to inject signals
        handlers that needs to in the context of the worker container.
        """
        return set()

    @classmethod
    def buildargs(cls) -> dict:
        """Build arguments for the built-in worker image."""
        return {
            "CELERY_VERSION": cls.version(),
            "CELERY_LOG_LEVEL": cls.log_level(),
            "CELERY_WORKER_NAME": cls.worker_name(),
            "CELERY_WORKER_QUEUE": cls.worker_queue(),
        }

    @classmethod
    def initial_env(cls, celery_worker_cluster_config: dict, initial: dict | None = None) -> dict:
        """Defines the environment variables for the worker container.

        See more: pytest_docker_tools.container()

        Args:
            celery_worker_cluster_config (dict): Environment variables to set.
            initial (dict | None, optional): Additional variables. Defaults to None.

        Returns:
            dict: Environment variables set for the worker container from the test case.
        """
        env = initial or {}
        env = {
            **DEFAULT_WORKER_ENV.copy(),
            **env,
        }

        config_mappings = [
            ("celery_broker_cluster_config", "CELERY_BROKER_URL"),
            ("celery_backend_cluster_config", "CELERY_RESULT_BACKEND"),
        ]

        for config_key, env_key in config_mappings:
            cluster_config = celery_worker_cluster_config.get(config_key)
            if cluster_config:
                env[env_key] = ";".join(cluster_config["urls"])
            else:
                del env[env_key]

        return env

    @classmethod
    def initial_content(
        cls,
        worker_tasks: set | None = None,
        worker_signals: set | None = None,
        worker_app: Celery | None = None,
        app_module: ModuleType | None = None,
        utils_module: ModuleType | None = None,
    ) -> dict:
        """Defines the initial content of the worker container.

        See more: pytest_docker_tools.volume()

        Args:
            worker_tasks (set | None, optional): Set of tasks modules. Defaults to None.
            worker_signals (set | None, optional): Set of signals handlers modules. Defaults to None.
            worker_app (Celery | None, optional): Celery app instance. Defaults to None.
            app_module (ModuleType | None, optional): app module. Defaults to None.
            utils_module (ModuleType | None, optional): utils module. Defaults to None.

        Returns:
            dict: Custom volume content for the worker container.
        """
        if app_module is None:
            app_module = cls.app_module()

        if utils_module is None:
            utils_module = cls.utils_module()

        if worker_tasks is None:
            worker_tasks = cls.tasks_modules()

        content = WorkerInitialContent()
        content.set_app_module(app_module)
        content.set_utils_module(utils_module)
        content.add_modules("tasks", worker_tasks)
        if worker_signals:
            content.add_modules("signals", worker_signals)
        if worker_app:
            content.set_app_name(worker_app.main)
            content.set_config_from_object(worker_app)

        return content.generate()

    @classmethod
    def ports(cls) -> dict | None:
        """Ports to expose from the worker container."""
        return DEFAULT_WORKER_PORTS



"""The pytest-celery plugin provides a set of built-in components called
:ref:`vendors`.

This module is part of the Redis Broker vendor.
"""

from __future__ import annotations

from pytest_celery.api.broker import CeleryTestBroker


class RedisTestBroker(CeleryTestBroker):
    pass



"""The pytest-celery plugin provides a set of built-in components called
:ref:`vendors`.

This module is part of the Redis Broker vendor.
"""

CELERY_REDIS_BROKER = "celery_redis_broker"
DEFAULT_REDIS_BROKER = "default_redis_broker"



"""The pytest-celery plugin provides a set of built-in components called
:ref:`vendors`.

This module is part of the RabbitMQ Broker vendor.
"""

CELERY_RABBITMQ_BROKER = "celery_rabbitmq_broker"
DEFAULT_RABBITMQ_BROKER = "default_rabbitmq_broker"
RABBITMQ_IMAGE = "rabbitmq:latest"
RABBITMQ_PORTS = {"5672/tcp": None}
RABBITMQ_ENV: dict = {}
RABBITMQ_CONTAINER_TIMEOUT = 120
RABBITMQ_PREFIX = "amqp://"



"""The pytest-celery plugin provides a set of built-in components called
:ref:`vendors`.

This module is part of the :ref:`built-in-worker` vendor.
"""

from __future__ import annotations

import time
from typing import Any
from typing import Iterable

import celery.utils
from celery import Task
from celery import shared_task


# ------------- add -------------
@shared_task
def add(x: int | float, y: int | float, z: int | float | None = None) -> int | float:
    """Pytest-celery internal task.

    This task adds two or three numbers together.

    Args:
        x (int | float): The first number.
        y (int | float): The second number.
        z (int | float | None, optional): The third number. Defaults to None.

    Returns:
        int | float: The sum of the numbers.
    """
    if z:
        return x + y + z
    else:
        return x + y


# ------------- add_replaced -------------
@shared_task(bind=True)
def add_replaced(
    self: Task,
    x: int | float,
    y: int | float,
    z: int | float | None = None,
    *,
    queue: str | None = None,
) -> None:
    """Pytest-celery internal task.

    This task replaces itself with the add task for the given arguments.

    Args:
        x (int | float): The first number.
        y (int | float): The second number.
        z (int | float | None, optional): The third number. Defaults to None.

    Raises:
        Ignore: Always raises Ignore.
    """
    queue = queue or "celery"
    raise self.replace(add.s(x, y, z).set(queue=queue))


# ------------- fail -------------
@shared_task
def fail(*args: tuple) -> None:
    """Pytest-celery internal task.

    This task raises a RuntimeError with the given arguments.

    Args:
        *args (tuple): Arguments to pass to the RuntimeError.

    Raises:
        RuntimeError: Always raises a RuntimeError.
    """
    args = (("Task expected to fail",) + args,)
    raise RuntimeError(*args)


# ------------- identity -------------
@shared_task
def identity(x: Any) -> Any:
    """Pytest-celery internal task.

    This task returns the input as is.

    Args:
        x (Any): Any value.

    Returns:
        Any: The input value.
    """
    return x


# ------------- noop -------------
@shared_task
def noop(*args: tuple, **kwargs: dict) -> None:
    """Pytest-celery internal task.

    This is a no-op task that does nothing.

    Returns:
        None: Always returns None.
    """
    return celery.utils.noop(*args, **kwargs)


# ------------- ping -------------
@shared_task
def ping() -> str:
    """Pytest-celery internal task.

    Used to check if the worker is up and running.

    Returns:
        str: Always returns "pong".
    """
    return "pong"


# ------------- sleep -------------
@shared_task
def sleep(seconds: float = 1, **kwargs: dict) -> bool:
    """Pytest-celery internal task.

    This task sleeps for the given number of seconds.

    Args:
        seconds (float, optional): The number of seconds to sleep. Defaults to 1.
        **kwargs (dict): Additional keyword arguments.

    Returns:
        bool: Always returns True.
    """
    time.sleep(seconds, **kwargs)
    return True


# ------------- xsum -------------
@shared_task
def xsum(nums: Iterable) -> int:
    """Pytest-celery internal task.

    This task sums a list of numbers, but also supports nested lists.

    Args:
        nums (Iterable): A list of numbers or nested lists.

    Returns:
        int: The sum of the numbers.
    """
    return sum(sum(num) if isinstance(num, Iterable) else num for num in nums)



"""The test setup represents the main entry point for accessing the celery
architecture from the test.

This module provides the base API for creating new test setups.
"""

from __future__ import annotations

from celery import Celery

from pytest_celery.api.backend import CeleryBackendCluster
from pytest_celery.api.backend import CeleryTestBackend
from pytest_celery.api.broker import CeleryBrokerCluster
from pytest_celery.api.broker import CeleryTestBroker
from pytest_celery.api.worker import CeleryTestWorker
from pytest_celery.api.worker import CeleryWorkerCluster
from pytest_celery.defaults import DEFAULT_WORKER_APP_NAME
from pytest_celery.defaults import RESULT_TIMEOUT
from pytest_celery.vendors.worker.tasks import ping


class CeleryTestSetup:
    """The test setup is the main entry point for accessing the celery
    architecture from the test. It is the glue that holds all of the relevant
    components of the specific test case environment.

    Each test case will have its own test setup instance, which is created for the
    test case by the plugin and is configured for the specific run and its given configurations.

    Responsibility Scope:
        Provide useful access to the celery architecture from the test.
    """

    def __init__(
        self,
        worker_cluster: CeleryWorkerCluster,
        broker_cluster: CeleryBrokerCluster,
        backend_cluster: CeleryBackendCluster,
        app: Celery = None,
    ):
        """Setup the base components of a setup.

        Args:
            worker_cluster (CeleryWorkerCluster): Precorfigured worker cluster.
            broker_cluster (CeleryBrokerCluster): Precorfigured broker cluster.
            backend_cluster (CeleryBackendCluster): Precorfigured backend cluster.
            app (Celery, optional): Celery app configured for all of the nodes. Defaults to None.
        """
        self._worker_cluster = worker_cluster
        self._broker_cluster = broker_cluster
        self._backend_cluster = backend_cluster
        self._app = app

        # Special internal ping task, does not conflict with user "ping" tasks
        self.ping = ping

    def __len__(self) -> int:
        """The total number of nodes in the setup."""
        nodes_count = 0
        if self.broker_cluster:
            nodes_count += len(self.broker_cluster)
        if self.backend_cluster:
            nodes_count += len(self.backend_cluster)
        if self.worker_cluster:
            nodes_count += len(self.worker_cluster)
        return nodes_count

    @property
    def app(self) -> Celery:
        """The celery app configured for all of the nodes."""
        return self._app

    @property
    def backend_cluster(self) -> CeleryBackendCluster | None:
        """The backend cluster of this setup."""
        return self._backend_cluster

    @property
    def backend(self) -> CeleryTestBackend | None:
        """The first backend node of the backend cluster."""
        return self._backend_cluster[0] if self._backend_cluster else None  # type: ignore

    @property
    def broker_cluster(self) -> CeleryBrokerCluster | None:
        """The broker cluster of this setup."""
        return self._broker_cluster

    @property
    def broker(self) -> CeleryTestBroker | None:
        """The first broker node of the broker cluster."""
        return self._broker_cluster[0] if self._broker_cluster else None  # type: ignore

    @property
    def worker_cluster(self) -> CeleryWorkerCluster | None:
        """The worker cluster of this setup."""
        return self._worker_cluster

    @property
    def worker(self) -> CeleryTestWorker | None:
        """The first worker node of the worker cluster."""
        return self._worker_cluster[0] if self._worker_cluster else None  # type: ignore

    @classmethod
    def name(cls) -> str:
        """The name of the setup."""
        # TODO: Possibly not needed/required refactoring
        return DEFAULT_WORKER_APP_NAME

    @classmethod
    def config(cls, celery_worker_cluster_config: dict) -> dict:
        """Creates a configuration dict to be used by app.config_from_object().
        The configuration is compiled from all of the nodes in the setup.

        Args:
            celery_worker_cluster_config (dict): The configuration of the worker cluster.

        Returns:
            dict: Celery-aware configuration dict.
        """
        if not celery_worker_cluster_config:
            raise ValueError("celery_worker_cluster_config is empty")

        celery_broker_cluster_config: dict = celery_worker_cluster_config.get("celery_broker_cluster_config", {})
        celery_backend_cluster_config: dict = celery_worker_cluster_config.get("celery_backend_cluster_config", {})
        config = {}
        if celery_broker_cluster_config:
            config["broker_url"] = ";".join(celery_broker_cluster_config["host_urls"])
        if celery_backend_cluster_config:
            config["result_backend"] = ";".join(celery_backend_cluster_config["host_urls"])
        return config

    @classmethod
    def update_app_config(cls, app: Celery) -> None:
        """Hook for updating the app configuration in a subclass.

        Args:
            app (Celery): App after initial configuration.
        """

    @classmethod
    def create_setup_app(cls, celery_setup_config: dict, celery_setup_app_name: str) -> Celery:
        """Creates a celery app for the setup.

        Args:
            celery_setup_config (dict): Celery configuration dict.
            celery_setup_app_name (str): Celery app name.

        Returns:
            Celery: Celery app configured for this setup.
        """
        if celery_setup_config is None:
            raise ValueError("celery_setup_config is None")

        if not celery_setup_app_name:
            raise ValueError("celery_setup_app_name is empty")

        app = Celery(celery_setup_app_name)
        app.config_from_object(celery_setup_config)
        cls.update_app_config(app)

        return app

    def teardown(self) -> None:
        """Teardown the setup."""

    def ready(self, ping: bool = False, control: bool = False, docker: bool = True) -> bool:
        """Check if the setup is ready for testing. All of the confirmations
        are optional.

        Warning:
            Enabling additional confirmations may hurt performance.
            Disabling all confirmations may result in false positive results.
            Use with caution.

        Args:
            ping (bool, optional): Confirm via ping task. Defaults to False.
            control (bool, optional): Confirm via ping control. Defaults to False.
            docker (bool, optional): Confirm via docker container status. Defaults to True.

        Returns:
            bool: True if the setup is ready for testing (all required confirmations passed).
        """
        ready = (
            self._is_task_ping_ready(ping) and self._is_control_ping_ready(control) and self._is_docker_ready(docker)
        )

        if ready:
            self._set_app_for_all_nodes()

        return ready

    def _is_docker_ready(self, docker: bool) -> bool:
        """Check if the node's containers are ready.

        Args:
            docker (bool): Flag to enable docker readiness check.

        Returns:
            bool: True if the node's containers are ready, False otherwise.
        """
        if not docker:
            return True

        return (
            (not self.broker_cluster or self.broker_cluster.ready())
            and (not self.backend_cluster or self.backend_cluster.ready())
            and (not self.worker_cluster or self.worker_cluster.ready())
        )

    def _is_control_ping_ready(self, control: bool) -> bool:
        """Check if worker nodes respond to control ping.

        Args:
            control (bool): Flag to enable control ping check.

        Returns:
            bool: True if control pings are successful, False otherwise.
        """
        if not control:
            return True

        responses = self.app.control.ping()
        return all([all([res["ok"] == "pong" for _, res in response.items()]) for response in responses])

    def _is_task_ping_ready(self, ping: bool) -> bool:
        """Check if worker nodes respond to ping task.

        Args:
            ping (bool): Flag to enable ping task check.

        Returns:
            bool: True if ping tasks are successful, False otherwise.
        """
        if not ping:
            return True

        worker: CeleryTestWorker
        for worker in self.worker_cluster:  # type: ignore
            res = self.ping.s().apply_async(queue=worker.worker_queue)
            if res.get(timeout=RESULT_TIMEOUT) != "pong":
                return False
        return True

    def _set_app_for_all_nodes(self) -> None:
        """Set the app instance for all nodes in the setup.

        This ensures each node has a reference to the centralized Celery
        app instance.
        """
        nodes: tuple = tuple()
        if self.broker_cluster:
            nodes += self.broker_cluster.nodes
        if self.backend_cluster:
            nodes += self.backend_cluster.nodes
        if self.worker_cluster:
            nodes += self.worker_cluster.nodes

        for node in nodes:
            node._app = self.app



"""See :ref:`vendors`."""


def _is_vendor_installed(vendor_name: str) -> bool:
    """Check if a vendor is installed.

    Args:
        vendor_name (str): Vendor package name.

    Returns:
        bool: True if the vendor is installed, False otherwise.
    """

    try:
        container_module = f"pytest_celery.vendors.{vendor_name}.container"
        __import__(container_module)
        return True
    except ImportError:
        return False



"""Pytest-celery entry point."""

# Backward compatibility with pytest-celery < 1.0.0
from celery.contrib.pytest import *  # noqa

# pytest-celery >= 1.0.0 infrastructure
from pytest_celery import *  # noqa



"""The pytest-celery plugin provides a set of built-in components called
:ref:`vendors`.

This module is part of the :ref:`built-in-worker` vendor.
"""

from __future__ import annotations

import inspect
from types import ModuleType
from typing import Any

from celery import Celery
from celery.app.base import PendingConfiguration

from pytest_celery.vendors.worker.defaults import DEFAULT_WORKER_APP_NAME


class WorkerInitialContent:
    """This class is responsible for generating the initial content for the
    worker container volume.

    Responsibility Scope:
        Prepare the worker container with the required filesystem, configurations and
        dependencies for your project.
    """

    class Parser:
        """Parser for the initial content of the worker container."""

        def imports_str(self, modules: set[ModuleType]) -> str:
            """Parse the given modules and return a string with the import
            statements.

            Args:
                modules (set[ModuleType]): A set of modules to parse.

            Returns:
                str: "from module import \\*" statements.
            """
            return "".join(f"from {module.__name__} import *\n" for module in modules)

        def imports_src(self, modules: set[ModuleType]) -> dict:
            """Parse the given modules and return a dict with the source code
            of the modules.

            Args:
                modules (set[ModuleType]): A set of modules to parse.

            Returns:
                dict: A dict with the source code of the modules.
            """
            src = dict()
            for module in modules:
                src[f"{module.__name__.replace('.', '/')}.py"] = inspect.getsource(module).encode()
            return src

        def app_name(self, name: str | None = None) -> str:
            """Generates the Celery app initialization string.

            Args:
                name (str | None, optional): The app name. Defaults to None.

            Returns:
                str: app = Celery(name)
            """
            name = name or DEFAULT_WORKER_APP_NAME
            return f"app = Celery('{name}')"

        def config(self, app: Celery | None = None) -> str:
            """Generates the Celery app configuration changes.

            Args:
                app (Celery | None, optional): Celery app with possibly changed config. Defaults to None.

            Raises:
                TypeError: If the app.conf.changes property is not a dict.

            Returns:
                str: config = {key: value, ...} or config = None
            """
            app = app or Celery(DEFAULT_WORKER_APP_NAME)

            # Accessing the app.conf.changes.data property will trigger the PendingConfiguration to be resolved
            # and the changes will be applied to the app.conf, so we make a clone app to avoid affecting the
            # original app object.
            tmp_app = Celery(app.main)
            tmp_app.conf = app.conf

            changes = tmp_app.conf.changes.copy()
            if isinstance(changes, PendingConfiguration):
                changes = changes.data.changes
            if not isinstance(changes, dict):
                raise TypeError(f"Unexpected type for app.conf.changes: {type(changes)}")
            del changes["deprecated_settings"]

            if changes:
                changes = (f"\t{repr(key)}: {repr(value)}" for key, value in changes.items())
                config = "config = {\n" + ",\n".join(changes) + "\n}" if changes else "config = None"
            else:
                config = "config = None"
            return config

    def __init__(
        self,
        app_module: ModuleType | None = None,
        utils_module: ModuleType | None = None,
    ) -> None:
        """Creates an initial content for the worker container. Defaults to
        built-in plugin-provided modules.

        Args:
            app_module (ModuleType | None, optional): App module adjusted for parsing. Defaults to None.
            utils_module (ModuleType | None, optional): Utils module with for running python code in the
            worker container. Defaults to None.
        """
        self.parser = self.Parser()
        self._initial_content = {
            "__init__.py": b"",
            "imports": dict(),  # Placeholder item
        }
        self.set_app_module(app_module)
        self.set_utils_module(utils_module)
        self.set_app_name()
        self.set_config_from_object()

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, WorkerInitialContent):
            return False
        try:
            return self.generate() == other.generate()
        except ValueError:
            return all(
                [
                    self._app_module_src == other._app_module_src,
                    self._utils_module_src == other._utils_module_src,
                    self._initial_content == other._initial_content,
                    self._app == other._app,
                    self._config == other._config,
                ]
            )

    def set_app_module(self, app_module: ModuleType | None = None) -> None:
        """Injects an app module into the initial content."""
        self._app_module_src: str | None

        if app_module:
            self._app_module_src = inspect.getsource(app_module)
        else:
            self._app_module_src = None

    def set_utils_module(self, utils_module: ModuleType | None = None) -> None:
        """Injects a utils module into the initial content."""
        self._utils_module_src: str | None

        if utils_module:
            self._utils_module_src = inspect.getsource(utils_module)
        else:
            self._utils_module_src = None

    def add_modules(self, name: str, modules: set[ModuleType]) -> None:
        """Adds a set of modules to the initial content.

        Args:
            name (str): Arbitrary unique name for the set of modules.
            modules (set[ModuleType]): A set of modules to add.
        """
        if not name:
            raise ValueError("name cannot be empty")

        if not modules:
            raise ValueError("modules cannot be empty")

        self._initial_content["imports"][name] = self.parser.imports_str(modules)  # type: ignore
        self._initial_content.update(self.parser.imports_src(modules))

    def set_app_name(self, name: str | None = None) -> None:
        """Sets the Celery app name.

        Args:
            name (str | None, optional): The app name. Defaults to None.
        """
        name = name or DEFAULT_WORKER_APP_NAME
        self._app = self.parser.app_name(name)

    def set_config_from_object(self, app: Celery | None = None) -> None:
        """Sets the Celery app configuration from the given app.

        Args:
            app (Celery | None, optional): Celery app with possibly changed config. Defaults to None.
        """
        self._config = self.parser.config(app)

    def generate(self) -> dict:
        """Generates the initial content for the worker container.

        Returns:
            dict: Initial content volume for the worker container.
        """
        initial_content = self._initial_content.copy()
        initial_content["app.py"] = self._generate_app_py(initial_content)
        initial_content["utils.py"] = self._generate_utils_py()
        return initial_content

    def _generate_app_py(self, initial_content: dict) -> bytes:
        """Generates the app.py file for the worker container.

        Args:
            initial_content (dict): The current initial content.

        Returns:
            bytes: The generated app.py file.
        """
        if not self._app_module_src:
            raise ValueError("Please use set_app_module() to define the app module before generating initial content")

        if not initial_content["imports"]:
            raise ValueError(
                "Please use set_utils_module() to define the utils module before generating initial content"
            )

        _imports: dict | Any = initial_content.pop("imports")
        imports = "{%s}" % "}{".join(_imports.keys())
        imports = imports.format(**_imports)

        app, config = self._app, self._config

        replacement_args = {
            "imports": "imports = None",
            "app": f'app = Celery("{DEFAULT_WORKER_APP_NAME}")',
            "config": "config = None",
        }
        self._app_module_src = self._app_module_src.replace(replacement_args["imports"], imports)
        self._app_module_src = self._app_module_src.replace(replacement_args["app"], app)
        self._app_module_src = self._app_module_src.replace(replacement_args["config"], config)

        return self._app_module_src.encode()

    def _generate_utils_py(self) -> bytes:
        """Generates the utils.py file for the worker container.

        Returns:
            bytes: The generated utils.py file.
        """
        if not self._utils_module_src:
            raise ValueError("Please set_utils_module() before generating initial content")

        return self._utils_module_src.encode()
